@@ -1,11 +1,10 @@
+import concurrent.futures
 from collections import namedtuple
 from enum import Enum
 
+from ktool.dyld import SymbolType
 from ktool.macho import Section
-from ktool.structs import objc2_class, objc2_class_t, objc2_class_ro_t, objc2_meth, objc2_meth_t, \
-    objc2_meth_list_entry_t, objc2_meth_list_t, objc2_prop_list_t, objc2_prop, objc2_prop_t, objc2_prot_list, \
-    objc2_prot_list_t, objc2_prot, objc2_prot_t, objc2_ivar_list, objc2_ivar_list_t, objc2_ivar, objc2_ivar_t, \
-    sizeof, struct
+from ktool.structs import *
 
 type_encodings = {
     "c": "char",
@@ -17,12 +16,13 @@ type_encodings = {
     "I": "unsigned int",
     "S": "unsigned short",
     "L": "unsigned long",
+    "A": "uint8_t",
     "Q": "NSUInteger",
     "f": "float",
     "d": "CGFloat",
     "b": "BOOL",
     "@": "id",
-    "B": "bool",
+    "B": "BOOL",
     "v": "void",
     "*": "char *",
     "#": "Class",
@@ -31,6 +31,96 @@ type_encodings = {
 }
 
 
+class ObjCLibrary:
+
+    def __init__(self, library, safe=False):
+        self.library = library
+        self.safe = safe
+        self.tp = TypeProcessor()
+        self.name = library.name
+
+        self.classlist = self._generate_classlist(None)
+        self.catlist = self._generate_catlist()
+        self.protolist = self._generate_protlist()
+
+    def _generate_catlist(self):
+        sect = None
+        for seg in self.library.segments:
+            for sec in self.library.segments[seg].sections:
+                if sec == "__objc_catlist":
+                    sect = self.library.segments[seg].sections[sec]
+
+        if not sect:
+            return []
+
+        cats = []  # meow
+        count = sect.size // 0x8
+        for offset in range(0, count):
+            cats.append(Category(self, sect.vm_address + offset * 0x8))
+
+        return cats
+
+    def _generate_classlist(self, classlimit):
+        sect = None
+        for seg in self.library.segments:
+            for sec in self.library.segments[seg].sections:
+                if sec == "__objc_classlist":
+                    sect = self.library.segments[seg].sections[sec]
+        # sect: Section = self.library.segments['__DATA_CONST'].sections['__objc_classlist']
+        if not sect:
+            return []
+        classes = []
+        cnt = sect.size // 0x8
+        for i in range(0, cnt):
+            if classlimit is None:
+                classes.append(Class(self, sect.vm_address + i * 0x8))
+            else:
+                oc = Class(self, sect.vm_address + i * 0x8)
+                if classlimit == oc.name:
+                    classes.append(oc)
+        return classes
+
+    # objc2_prot_list = namedtuple("objc2_prot_list", ["off", "cnt"])
+    # objc2_prot_list_t = struct(objc2_prot_list, [8])
+
+    def _generate_protlist(self):
+
+        sect = None
+        for seg in self.library.segments:
+            for sec in self.library.segments[seg].sections:
+                if sec == "__objc_protolist":
+                    sect = self.library.segments[seg].sections[sec]
+        # sect: Section = self.library.segments['__DATA_CONST'].sections['__objc_classlist']
+        if not sect:
+            return []
+
+        protos = []
+
+        cnt = sect.size // 0x8
+        for i in range(0, cnt):
+            ptr = sect.vm_address + i * 0x8
+            loc = self.library.get_bytes(ptr, 0x8, vm=True)
+            proto = self.library.load_struct(loc, objc2_prot_t, vm=True)
+            try:
+                protos.append(Protocol(self, proto, loc))
+            except:
+                pass
+
+        return protos
+
+    def get_bytes(self, offset: int, length: int, vm=False, sectname=None):
+        return self.library.get_bytes(offset, length, vm, sectname)
+
+    def load_struct(self, addr: int, struct_type: struct, vm=True, sectname=None, endian="little"):
+        return self.library.load_struct(addr, struct_type, vm, sectname, endian)
+
+    def get_str_at(self, addr: int, count: int, vm=True, sectname=None):
+        return self.library.get_str_at(addr, count, vm, sectname)
+
+    def get_cstr_at(self, addr: int, limit: int = 0, vm=True, sectname=None):
+        return self.library.get_cstr_at(addr, limit, vm, sectname)
+
+
 class Struct:
     def __init__(self, processor, type_str: str):
         # {name=dd}
@@ -38,6 +128,10 @@ class Struct:
         # Remove the outer {}, then get everything to the left of the equal sign
         self.name = type_str[1:-1].split('=')[0]
 
+        if '=' not in type_str:
+            self.fields = []
+            return
+
         # Remove the outer {},
         # get everything after the first = sign,
         # Process that via the processor
@@ -124,6 +218,7 @@ class TypeProcessor:
                     pc = 0
             return types
         except:
+            # TODO: Fix type recognition issues to best of abilities and then remove this assertion
             raise AssertionError(type)
 
     @staticmethod
@@ -174,6 +269,8 @@ class Ivar:
 
     def __str__(self):
         ret = ""
+        if self.type.startswith('<'):
+            ret += "id"
         ret += self.type + ' '
         if self.is_id:
             ret += '*'
@@ -193,16 +290,19 @@ class Ivar:
 
 
 class Method:
-    def __init__(self, library, objc_class, method: objc2_meth, vmaddr: int):
-        self.objc_class = objc_class
-
+    def __init__(self, library, meta, method: objc2_meth, vmaddr: int):
+        self.meta = meta
         try:
             self.sel = library.get_cstr_at(method.selector, 0, vm=True, sectname="__objc_methname")
             typestr = library.get_cstr_at(method.types, 0, vm=True, sectname="__objc_methtype")
         except ValueError as ex:
-            selref = library.get_bytes(method.selector + vmaddr, 8, vm=True)
-            self.sel = library.get_cstr_at(selref, 0, vm=True, sectname="__objc_methname")
-            typestr = library.get_cstr_at(method.types + vmaddr + 4, 0, vm=True, sectname="__objc_methtype")
+            try:
+                selref = library.get_bytes(method.selector + vmaddr, 8, vm=True)
+                self.sel = library.get_cstr_at(selref, 0, vm=True, sectname="__objc_methname")
+                typestr = library.get_cstr_at(method.types + vmaddr + 4, 0, vm=True, sectname="__objc_methtype")
+            except ValueError as ex:
+                self.sel = library.get_cstr_at(method.selector + vmaddr, 0, vm=True, sectname="__objc_methname")
+                typestr = library.get_cstr_at(method.types + vmaddr + 4, 0, vm=True, sectname="__objc_methtype")
         except Exception as ex:
             raise ex
 
@@ -232,7 +332,7 @@ class Method:
             return ptraddon + type.value.name
 
     def _build_method_signature(self):
-        dash = "+" if self.objc_class.meta else "-"
+        dash = "+" if self.meta else "-"
         ret = "(" + self.return_string + ")"
 
         if len(self.arguments) == 0:
@@ -252,41 +352,6 @@ class Method:
         return dash + ret + sig
 
 
-class ObjCLibrary:
-
-    def __init__(self, library):
-        self.library = library
-        self.tp = TypeProcessor()
-        self.name = library.name
-
-        self.classlist = self._generate_classlist(None)
-
-    def _generate_classlist(self, classlimit):
-        sect: Section = self.library.segments['__DATA_CONST'].sections['__objc_classlist']
-        classes = []
-        cnt = sect.size // 0x8
-        for i in range(0, cnt):
-            if classlimit is None:
-                classes.append(Class(self, sect.vm_address + i * 0x8))
-            else:
-                oc = Class(self, sect.vm_address + i * 0x8)
-                if classlimit == oc.name:
-                    classes.append(oc)
-        return classes
-
-    def get_bytes(self, offset: int, length: int, vm=False, sectname=None):
-        return self.library.get_bytes(offset, length, vm, sectname)
-
-    def load_struct(self, addr: int, struct_type: struct, vm=True, sectname=None, endian="little"):
-        return self.library.load_struct(addr, struct_type, vm, sectname, endian)
-
-    def get_str_at(self, addr: int, count: int, vm=True, sectname=None):
-        return self.library.get_str_at(addr, count, vm, sectname)
-
-    def get_cstr_at(self, addr: int, limit: int = 0, vm=True, sectname=None):
-        return self.library.get_cstr_at(addr, limit, vm, sectname)
-
-
 class LinkedClass:
     def __init__(self, classname, libname):
         self.classname = classname
@@ -315,6 +380,9 @@ class Class:
         self.superclass = ""
         self.linkedlibs = []
         self.linked_classes = []
+        self.fdec_classes = []
+        self.fdec_prots = []
+        self.struct_list = []
         # Classes imported in this class from the same mach-o
         if not objc2class:
             self.objc2_class: objc2_class = self._load_objc2_class(ptr)
@@ -354,18 +422,14 @@ class Class:
             # Linked Superclass
             struct_size = sizeof(objc2_class_t)
             struct_location = objc2_class_item.off
-            for symbol in self.library.library.binding_table.symbol_table:
-                try:
-                    action_file_location = self.library.library.vm.get_file_address(symbol.addr)
-                except ValueError:
-                    continue
-                if action_file_location == struct_location + 0x8:
-                    try:
-                        self.superclass = symbol.name[1:]
-                        self.linked_classes.append(LinkedClass(symbol.name[1:], self.library.library.linked[int(symbol.ordinal) - 1].install_name))
-                    except IndexError:
-                        continue
-                    break
+            symbol = self.library.library.binding_table.lookup_table[objc2_class_location + 8]
+            self.superclass = symbol.name[1:]
+            try:
+                self.linked_classes.append(LinkedClass(symbol.name[1:], self.library.library.linked[
+                    int(symbol.ordinal) - 1].install_name))
+            except IndexError:
+                # sometimes load commands seem to do a fucky wucky
+                pass
         if objc2_class_item.isa != 0 and objc2_class_item.isa <= 0xFFFFFFFFFF and not self.meta:
             try:
                 metaclass_item: objc2_class = self.library.load_struct(objc2_class_item.isa, objc2_class_t)
@@ -405,9 +469,14 @@ class Class:
             else:
                 meth = self.library.load_struct(ea, objc2_meth_t, vm=False)
             try:
-                methods.append(Method(self.library, self, meth, vm_ea))
+                method = Method(self.library, self.meta, meth, vm_ea)
+                methods.append(method)
+                for type in method.types:
+                    if type.type == EncodedType.STRUCT:
+                        self.struct_list.append(type.value)
+
             except Exception as ex:
-                continue
+                pass
             if uses_relative_methods:
                 ea += sizeof(objc2_meth_list_entry_t)
                 vm_ea += sizeof(objc2_meth_list_entry_t)
@@ -433,10 +502,14 @@ class Class:
         for i in range(1, proplist_head.count + 1):
             prop = self.library.load_struct(ea, objc2_prop_t, vm=False)
             try:
-                properties.append(Property(self.library, self, prop, vm_ea))
+                property = Property(self.library, prop, vm_ea)
+                properties.append(property)
+                if hasattr(property, 'attr'):
+                    if property.attr.type.type == EncodedType.STRUCT:
+                        self.struct_list.append(property.attr.type.value)
             except ValueError as ex:
                 # continue
-                raise ex
+                pass
             ea += sizeof(objc2_prop_t)
             vm_ea += sizeof(objc2_prop_t)
 
@@ -451,7 +524,10 @@ class Class:
         for i in range(1, protlist.cnt + 1):
             prot_loc = self.library.get_bytes(ea + i * 8, 8, vm=False)
             prot = self.library.load_struct(prot_loc, objc2_prot_t, vm=True)
-            prots.append(Protocol(self.library, self, prot, prot_loc))
+            try:
+                prots.append(Protocol(self.library, self, prot, prot_loc))
+            except Exception as ex:
+                continue
         return prots
 
     def _process_ivars(self):
@@ -463,13 +539,18 @@ class Class:
         for i in range(1, ivarlist.cnt + 1):
             ivar_loc = ea + sizeof(objc2_ivar_t) * (i - 1)
             ivar = self.library.load_struct(ivar_loc, objc2_ivar_t, vm=False)
-            ivars.append(Ivar(self.library, self, ivar, ivar_loc))
+            try:
+                ivar_object = Ivar(self.library, self, ivar, ivar_loc)
+                ivars.append(ivar_object)
+            except Exception as ex:
+                continue
         return ivars
 
 
 attr_encodings = {
     "&": "retain",
     "N": "nonatomic",
+    "W": "__weak",
     "R": "readonly",
     "C": "copy"
 }
@@ -477,7 +558,7 @@ property_attr = namedtuple("property_attr", ["type", "attributes", "ivar", "is_i
 
 
 class Property:
-    def __init__(self, library, objc_class, property: objc2_prop, vmaddr: int):
+    def __init__(self, library, property: objc2_prop, vmaddr: int):
         self.library = library
         self.property = property
 
@@ -501,6 +582,8 @@ class Property:
         if len(self.attributes) > 0:
             ret += '(' + ', '.join(self.attributes) + ') '
 
+        if self.type.startswith('<'):
+            ret += "id"
         ret += self.type + ' '
 
         if self.is_id:
@@ -546,9 +629,169 @@ class Property:
         return property_attr(ptype, property_attributes, ivar, is_id, type_str)
 
 
+class Category:
+    def __init__(self, library, ptr):
+        self.library = library
+        self.ptr = ptr
+        loc = self.library.get_bytes(ptr, 8, vm=True)
+
+        self.struct: objc2_category = self.library.load_struct(loc, objc2_category_t, vm=True)
+        self.name = self.library.get_cstr_at(self.struct.name, vm=True)
+        self.classname = ""
+        try:
+            sym = self.library.library.binding_table.lookup_table[loc + 8]
+            self.classname = sym.name[1:]
+        except:
+            pass
+
+        instmeths = self._process_methods(self.struct.inst_meths)
+        classmeths = self._process_methods(self.struct.class_meths, True)
+
+        self.methods = instmeths + classmeths
+        self.properties = self._process_props(self.struct.props)
+        self.protocols = []
+
+    def _process_methods(self, loc, meta=False):
+        methods = []
+
+        if loc == 0:
+            return methods  # Useless Subclass
+
+        vm_ea = loc
+        methlist_head = self.library.load_struct(loc, objc2_meth_list_t)
+        ea = methlist_head.off
+
+        # https://github.com/arandomdev/DyldExtractor/blob/master/DyldExtractor/objc/objc_structs.py#L79
+        RELATIVE_METHODS_SELECTORS_ARE_DIRECT_FLAG = 0x40000000
+        RELATIVE_METHOD_FLAG = 0x80000000
+        METHOD_LIST_FLAGS_MASK = 0xFFFF0000
+
+        uses_relative_methods = methlist_head.entrysize & METHOD_LIST_FLAGS_MASK != 0
+
+        ea += 8
+        vm_ea += 8
+        for i in range(1, methlist_head.count + 1):
+            if uses_relative_methods:
+                meth = self.library.load_struct(ea, objc2_meth_list_entry_t, vm=False)
+            else:
+                meth = self.library.load_struct(ea, objc2_meth_t, vm=False)
+            try:
+                methods.append(Method(self.library, meta, meth, vm_ea))
+            except Exception as ex:
+                pass
+            if uses_relative_methods:
+                ea += sizeof(objc2_meth_list_entry_t)
+                vm_ea += sizeof(objc2_meth_list_entry_t)
+            else:
+                ea += sizeof(objc2_meth_t)
+                vm_ea += sizeof(objc2_meth_t)
+
+        return methods
+
+    def _process_props(self, location):
+        properties = []
+
+        if location == 0:
+            return properties
+
+        vm_ea = location
+        proplist_head = self.library.load_struct(location, objc2_prop_list_t)
+
+        ea = proplist_head.off
+        ea += 8
+        vm_ea += 8
+
+        for i in range(1, proplist_head.count + 1):
+            prop = self.library.load_struct(ea, objc2_prop_t, vm=False)
+            try:
+                properties.append(Property(self.library, prop, vm_ea))
+            except ValueError as ex:
+                # continue
+                pass
+            ea += sizeof(objc2_prop_t)
+            vm_ea += sizeof(objc2_prop_t)
+
+        return properties
+
+
+# objc2_prot = namedtuple("objc2_prot", ["off", "isa", "name", "prots", "inst_meths", "class_meths",
+# "opt_inst_meths", "opt_class_meths", "inst_props", "cb", "flags"])
+# objc2_prot_t = struct(objc2_prot, [8, 8, 8, 8, 8, 8, 8, 8, 4, 4])
+
 class Protocol:
-    def __init__(self, library, objc_class, protocol: objc2_prot, vmaddr: int):
+    def __init__(self, library, protocol: objc2_prot, vmaddr: int):
+        self.library = library
         self.name = library.get_cstr_at(protocol.name, 0, vm=True)
 
+        self.methods = self._process_methods(protocol.inst_meths)
+        self.methods += self._process_methods(protocol.class_meths, True)
+
+        self.opt_methods = self._process_methods(protocol.opt_inst_meths)
+        self.opt_methods += self._process_methods(protocol.opt_class_meths, True)
+
+        self.properties = self._process_props(protocol.inst_props)
+
+    def _process_methods(self, loc, meta=False):
+        methods = []
+
+        if loc == 0:
+            return methods  # Useless Subclass
+
+        vm_ea = loc
+        methlist_head = self.library.load_struct(loc, objc2_meth_list_t)
+        ea = methlist_head.off
+
+        # https://github.com/arandomdev/DyldExtractor/blob/master/DyldExtractor/objc/objc_structs.py#L79
+        RELATIVE_METHODS_SELECTORS_ARE_DIRECT_FLAG = 0x40000000
+        RELATIVE_METHOD_FLAG = 0x80000000
+        METHOD_LIST_FLAGS_MASK = 0xFFFF0000
+
+        uses_relative_methods = methlist_head.entrysize & METHOD_LIST_FLAGS_MASK != 0
+
+        ea += 8
+        vm_ea += 8
+        for i in range(1, methlist_head.count + 1):
+            if uses_relative_methods:
+                meth = self.library.load_struct(ea, objc2_meth_list_entry_t, vm=False)
+            else:
+                meth = self.library.load_struct(ea, objc2_meth_t, vm=False)
+            try:
+                methods.append(Method(self.library, meta, meth, vm_ea))
+            except Exception as ex:
+                pass
+            if uses_relative_methods:
+                ea += sizeof(objc2_meth_list_entry_t)
+                vm_ea += sizeof(objc2_meth_list_entry_t)
+            else:
+                ea += sizeof(objc2_meth_t)
+                vm_ea += sizeof(objc2_meth_t)
+
+        return methods
+
+    def _process_props(self, location):
+        properties = []
+
+        if location == 0:
+            return properties
+
+        vm_ea = location
+        proplist_head = self.library.load_struct(location, objc2_prop_list_t)
+
+        ea = proplist_head.off
+        ea += 8
+        vm_ea += 8
+
+        for i in range(1, proplist_head.count + 1):
+            prop = self.library.load_struct(ea, objc2_prop_t, vm=False)
+            try:
+                properties.append(Property(self.library, prop, vm_ea))
+            except ValueError as ex:
+                # continue
+                pass
+            ea += sizeof(objc2_prop_t)
+            vm_ea += sizeof(objc2_prop_t)
+
+        return properties
+
     def __str__(self):
         return self.name
